# '{{ project_name }}' NBX-Jobs

This is an autogenerated README file for [NBX-Jobs](www.nimblebox.ai/jobs) which was:
* created at: {{ created_time }}
* name: {{ project_name }}
* created by: {{ username }}

Steps to complete:

1. Add your code/folders/csv/etc. in this folder and optionally requirements.txt
2. Fill The Three functions (`get_operator`, `deploy`, `run`) in `exe.py` and `nboxcli jobs deploy .`
3. Sit back. Relax. And strap on your seat belt. - Dr. Dre

Rule of Thumb:

1. If you are scheduling, your `op` forward should not takes arguments, otherwise
you are executing the same process over and over, which is not what you want.
If you still want to do so, try moving as much as possible to `__init__`.
2. If you are running this one-time, then you should define arguments in `run` function.
